@@ -1,6 +1,7 @@
 # -*- coding: utf-8 -*-
-#   Copyright (C) 2008-2009, 2012-2013, 2015-2016, 2018, 2020
-#   Rocky Bernstein <rocky@gnu.org>
+#
+#   Copyright (C) 2008-2009, 2012-2013, 2015-2016, 2018, 2020-2021,
+#   2023-2025 Rocky Bernstein <rocky@gnu.org>
 #
 #   This program is free software: you can redistribute it and/or modify
 #   it under the terms of the GNU General Public License as published by
@@ -53,25 +54,29 @@ Synopsis
 
 """
 
-import hashlib, linecache, os, re, sys
+import hashlib
+import linecache
+import os
 import os.path as osp
-
+import re
+import sys
 from collections import namedtuple
+from importlib.util import find_spec, source_from_cache
+from typing import List, Optional
+from term_background import is_dark_background
 
 from pygments import highlight
+from pygments.formatters import Terminal256Formatter, TerminalFormatter
 from pygments.lexers import PythonLexer
-from pygments.formatters import TerminalFormatter, Terminal256Formatter
+from xdis.lineoffsets import lineoffsets_in_file
+from xdis.version_info import PYTHON3
 
-from xdis import PYTHON3, PYTHON_VERSION, lineoffsets_in_file
+from pyficache.pyasm import PyasmLexer
 from pyficache.line_numbers import code_linenumbers_in_file
 
 PYVER = "%s%s" % sys.version_info[0:2]
 
-if PYTHON3:
-    large_int = sys.maxsize
-else:
-    large_int = sys.maxint
-
+large_int = sys.maxsize
 
 default_opts = {
     "reload_on_change": False,  # Check if file has changed since last
@@ -84,26 +89,26 @@ default_opts = {
 }
 
 
+def grep_first_line(lines: List[str], pattern) -> Optional[str]:
+    """
+    Greps for the first line in "lines" that matches pattern "pattern".
+    """
+    matching_lines = [line for line in lines if re.search(pattern, line)]
+    return matching_lines[0] if len(matching_lines) > 0 else None
+
+
 def get_option(key, options):
     global default_opts
     if not options or key not in options:
         return default_opts.get(key)
-    else:
-        return options[key]
-    return None  # Not reached
+    return options[key]
 
 
-def has_trailing_nl(string):
+def has_trailing_nl(string) -> bool:
     return len(string) > 0 and "\n" == string[-1]
 
 
-if PYTHON_VERSION >= 3.4:
-    from importlib.util import source_from_cache, resolve_name, find_spec
-else:
-    source_from_cache = resolve_name = find_spec = None
-
-
-def resolve_name_to_path(path_or_name):
+def resolve_name_to_path(path_or_name: str) -> str:
     """Try to "resolve" `path_or_name` info its constituent file path.
 
     `path_or_name` could be either a
@@ -124,7 +129,7 @@ def resolve_name_to_path(path_or_name):
     if source_from_cache:
         try:
             source_path = source_from_cache(path_or_name)
-        except:
+        except Exception:
             pass
         else:
             if source_path:
@@ -134,7 +139,7 @@ def resolve_name_to_path(path_or_name):
     if find_spec:
         try:
             spec = find_spec(path_or_name)
-        except:
+        except Exception:
             spec = None
         else:
             if spec and spec.origin:
@@ -144,7 +149,7 @@ def resolve_name_to_path(path_or_name):
     if re.match(".*py[co]$", path_or_name):
         if PYTHON3:
             return re.sub(
-                r"(.*)__pycache__/(.+)\.cpython-%s.py[co]$" % PYVER,
+                rf"(.*)__pycache__/(.+)\.cpython-{PYVER}.py[co]$",
                 "\\1\\2.py",
                 path_or_name,
             )
@@ -156,11 +161,22 @@ def resolve_name_to_path(path_or_name):
 
 
 class LineCacheInfo:
-    def __init__(self, stat, line_numbers, linestarts, lines, path, sha1, eols=None):
+    def __init__(
+        self,
+        stat,
+        line_numbers,
+        linestarts,
+        lines,
+        path,
+        sha1,
+        eols=None,
+        code_map=None,
+    ):
         self.stat, self.lines, self.path, self.sha1 = (stat, lines, path, sha1)
         self.line_numbers = line_numbers
         self.linestarts = linestarts
         self.eols = eols
+        self.code_map = code_map
         return
 
     pass
@@ -170,6 +186,7 @@ class LineCacheInfo:
 # or __file__. The value is a LineCacheInfo object.
 file_cache = {}
 script_cache = {}
+pyasm_files = set()
 
 # `file2file_remap` maps a path (a string) to another path key in file_cache (a
 # string).
@@ -368,9 +385,7 @@ def cache_file(filename, reload_on_change=False, opts=default_opts):
         pass
     if filename in file_cache:
         return file_cache[filename].path
-    else:
-        return None
-    return  # Not reached
+    return None
 
 
 def is_cached(file_or_script):
@@ -401,20 +416,29 @@ def getline(file_or_script, line_number, opts=default_opts):
     """
     filename = unmap_file(file_or_script)
     filename, line_number = unmap_file_line(filename, line_number)
+
+    is_pyasm = opts.get("is_pyasm", filename.endswith(".pyasm"))
     lines = getlines(filename, opts)
-    if lines and line_number >= 1 and line_number <= maxline(filename):
+    if is_pyasm:
+        lines = getlines(filename, {"output": "plain"})
+        fmt = opts.get("output", "plain")
+        line = grep_first_line(lines, f"^[ ]+{line_number}:")
+        if fmt == "plain":
+            return line
+        else:
+            return highlight_string(line, fmt)
+
+    elif lines and line_number >= 1 and line_number <= maxline(filename):
         line = lines[line_number - 1]
         if get_option("strip_nl", opts):
             return line.rstrip("\n")
         else:
             return line
         pass
-    else:
-        return None
-    return  # Not reached
+    return None
 
 
-def getlines(filename, opts=default_opts):
+def getlines(filename, opts=default_opts, is_pyasm: Optional[bool] = None):
     """Read lines of *filename* and cache the results. However, if
     *filename* was previously cached use the results from the
     cache. Return *None* if we can not get lines
@@ -422,14 +446,25 @@ def getlines(filename, opts=default_opts):
     if get_option("reload_on_change", opts):
         checkcache(filename)
     fmt = get_option("output", opts)
+    if fmt == "plain":
+        cs = "plain"
+    else:
+        cs = opts.get("style")
     highlight_opts = {"bg": fmt}
-    cs = opts.get("style")
 
-    # Colorstyle of Terminal255Formatter takes precidence over
-    # light/dark colorthemes of TerminalFormatter
+    if is_pyasm is None:
+        is_pyasm = filename.endswith(".pyasm")
+
+    # Set list style based on "style" option passed
+    # if no style given use "monokai" for dark backgrounds,
+    # and "tango" for light backgrounds.
     if cs:
         highlight_opts["style"] = cs
         fmt = cs
+    elif is_dark_background:
+        highlight_opts["style"] = "monokai"
+    else:
+        highlight_opts["style"] = "tango"
 
     if filename not in file_cache:
         update_cache(filename, opts)
@@ -438,6 +473,9 @@ def getlines(filename, opts=default_opts):
             return None
         pass
     lines = file_cache[filename].lines
+    if is_pyasm:
+        # Until we has a pyasm lexer.
+        highlight_opts["lexer"] = pyasm_lexer
     if fmt not in lines.keys():
         lines[fmt] = highlight_array(lines["plain"], **highlight_opts)
         pass
@@ -452,6 +490,7 @@ def highlight_array(array, trailing_nl=True, bg="light", **options):
     return lines
 
 
+pyasm_lexer = PyasmLexer()
 python_lexer = PythonLexer()
 
 # TerminalFormatter uses a colorTHEME with light and dark pairs.
@@ -461,18 +500,21 @@ light_terminal_formatter = TerminalFormatter(bg="light")
 terminal_256_formatter = Terminal256Formatter()
 
 
-def highlight_string(string, bg="light", **options):
+def highlight_string(string, bg="light", **options) -> str:
     global terminal_256_formatter
+    if "lexer" in options:
+        lexer = options.pop("lexer")
+    else:
+        lexer = python_lexer
     if options.get("style"):
         if terminal_256_formatter.style != options["style"]:
             terminal_256_formatter = Terminal256Formatter(style=options["style"])
             del options["style"]
-        return highlight(string, python_lexer, terminal_256_formatter, **options)
+        return highlight(string, lexer, terminal_256_formatter, **options)
     elif "light" == bg:
-        return highlight(string, python_lexer, light_terminal_formatter, **options)
+        return highlight(string, lexer, light_terminal_formatter, **options)
     else:
-        return highlight(string, python_lexer, dark_terminal_formatter, **options)
-    pass
+        return highlight(string, lexer, dark_terminal_formatter, **options)
 
 
 def path(filename):
@@ -483,15 +525,39 @@ def path(filename):
     return file_cache[filename].path
 
 
-def remap_file(from_file, to_file):
+def remap_file(from_file: str, to_file: str, is_pyasm: bool = False):
     """Make *to_file* be a synonym for *from_file*"""
     file2file_remap[to_file] = from_file
+    if is_pyasm:
+        pyasm_files.add(from_file)
     return
 
 
+# Hash to remap filename by regular expression.
+# For example, often we may want to remap the beginning of a path to
+# something else because it may be mounted, so the prefix changes.
+remap_re_hash = {}
+
+
+def add_remap_pat(pat, replace, clear_remap=True):
+    global file2file_remap
+    remap_re_hash[re.compile(pat)] = (pat, replace)
+    if clear_remap:
+        file2file_remap = {}
+
+
+def remap_file_pat(from_file, remap_re_hash=remap_re_hash):
+    """If *from_file* matches remap_patterns do the replacement"""
+    for pat, replace_tup in remap_re_hash.items():
+        match = pat.match(from_file)
+        if match:
+            return re.sub(pat, replace_tup[1], from_file)
+    return from_file
+
+
 def remap_file_lines(from_path, to_path, line_map_list):
     """Adds line_map list to the list of association of from_file to
-       to to_file"""
+    to to_file"""
     from_path = resolve_name_to_path(from_path)
     cache_file(to_path)
     remap_entry = file2file_remap_lines.get(to_path)
@@ -626,6 +692,7 @@ def cache_code_lines(
         code_info = lineoffsets_in_file(fullname, toplevel_only=toplevel_only)
         file_info.line_numbers = code_info.line_numbers(include_offsets=include_offsets)
         file_info.linestarts = code_info.linestarts
+        file_info.code_map = code_info.code_map
         pass
     return file_info
 
@@ -636,38 +703,44 @@ def code_lines(
     """Return the line numbers, bytecode offsets, and code object that are
     (or would be) stored in the bytecode for `filename`.
     """
-    file_info = cache_code_lines(filename, toplevel_only, include_offsets)
+    file_info = cache_code_lines(
+        filename, toplevel_only=toplevel_only, include_offsets=include_offsets
+    )
     if not file_info:
         return None
     return file_info
 
 
 def code_line_info(
-    filename,
-    line_number,
+    filename: str,
+    line_number: int,
     reload_on_change=False,
     toplevel_only=False,
     include_offsets=True,
-):
+) -> tuple:
     """Return the bytecode information that is associated with
     `line_number` in the bytecode for `filename`.
     """
-    file_info = cache_code_lines(filename,
-                                 reload_on_change=reload_on_change,
-                                 toplevel_only=toplevel_only,
-                                 include_offsets=include_offsets)
+    file_info = cache_code_lines(
+        filename,
+        reload_on_change=reload_on_change,
+        toplevel_only=toplevel_only,
+        include_offsets=include_offsets,
+    )
     if not file_info:
         return None
-    return file_info.line_numbers.get(line_number, None)
+    return file_info.code_map, file_info.line_numbers.get(line_number, None)
 
 
 def code_offset_info(
-    filename, offset, reload_on_change=False,
+    filename,
+    offset,
+    reload_on_change=False,
 ):
     """Return the bytecode information that is associated with
     `offset` in the bytecode for `filename`.
 
-    Each entry is a line number list of instruction offests associated
+    Each entry is a line number list of instruction offsets associated
     with that line number and a code object where this can be found.
 
     This comes from findlinestarts() from xdis which is the same
@@ -694,8 +767,20 @@ def is_mapped_file(filename):
 
 
 def unmap_file(filename):
-    # FIXME: this is wrong?
-    return file2file_remap.get(filename, filename)
+    # If it is in the cache, use that.
+    if filename in file2file_remap:
+        return file2file_remap[filename]
+
+    # If there is a pattern remapping, do the
+    # remapping and cache the results
+    if remap_re_hash:
+        unmapped_file = remap_file_pat(filename)
+        if unmapped_file != filename:
+            remap_file(filename, unmapped_file)
+            return unmapped_file
+
+    # No remappings done, so return what was given
+    return filename
 
 
 def unmap_file_line(filename, line_number, reverse=False):
@@ -725,6 +810,8 @@ def unmap_file_line(filename, line_number, reverse=False):
                 break
             last_t = t
         pass
+    elif filename in pyasm_files:
+        pass
     return (filename, mapped_line_number)
 
 
@@ -764,7 +851,7 @@ def update_cache(filename, opts=default_opts, module_globals=None):
                         path=path,
                         sha1=None,
                     )
-                except:
+                except Exception:
                     pass
                 pass
             if orig_filename != filename:
@@ -817,7 +904,7 @@ def update_cache(filename, opts=default_opts, module_globals=None):
         # Try looking through the module search path, which is only useful
         # when handling a relative filename.
         stat = None
-        for dirname in sys.path:
+        for dirname in ["."] + sys.path:
             path = osp.join(dirname, filename)
             if osp.exists(path):
                 stat = os.stat(path)
@@ -832,7 +919,7 @@ def update_cache(filename, opts=default_opts, module_globals=None):
         with open(path, mode) as fp:
             lines = {"plain": fp.readlines()}
             eols = fp.newlines
-    except:
+    except Exception:
         return None
 
     # FIXME: DRY with code above
@@ -860,6 +947,7 @@ def update_cache(filename, opts=default_opts, module_globals=None):
         path=path,
         sha1=None,
         eols=eols,
+        code_map={},
     )
     file2file_remap[path] = filename
     return True
@@ -867,7 +955,6 @@ def update_cache(filename, opts=default_opts, module_globals=None):
 
 # example usage
 if __name__ == "__main__":
-
     z = lambda x, y: x + y
 
     def yes_no(var):
@@ -877,12 +964,20 @@ if __name__ == "__main__":
         prefix2 = "not "
         if var:
             return prefix1
-        else:
-            return prefix2
-        return  # Not reached
+        return prefix2
+
+    remap_file("../test/seven-313.pyasm", "seven.py", is_pyasm=True)
+    line = getline("seven.py", 4)
+    print(line)
+    line = getline("seven.py", 4, {"output": "light"})
+    print(line)
+    add_remap_pat("^/code", "/tmp/project")
+    print(remap_file_pat("/code/setup.py"))
 
     # print(resolve_name_to_path("os"))
-    # print(getline(__file__, 1, {"output": "dark"}))
+    print(
+        getline(__file__, 1, {"style": "monokai" if is_dark_background() else "tango"})
+    )
     # print(getline(__file__, 2, {"output": "light"}))
     # from pygments.styles import STYLE_MAP
 
@@ -910,15 +1005,18 @@ if __name__ == "__main__":
 
     update_cache(__file__)
     checkcache(__file__)
-    print("%s has %s lines" % (__file__, size(__file__)))
-    print("%s code_lines data:\n" % __file__)
+    print(f"{__file__} has {size(__file__)} lines")
+    print(f"{__file__} code_lines data:\n")
 
     line_info = code_offset_info(__file__, 0)
-    print("Starting line for file (bytecode offset 0) is %s" % line_info)
+    print(f"Starting line for file (bytecode offset 0) is {line_info}")
     line_info = code_lines(__file__).line_numbers
     for line_num, li in line_info.items():
         print("\tline: %4d: %s" % (line_num, ", ".join([str(i.offsets) for i in li])))
     print("=" * 30)
+    for mod, code in file_cache[__file__].code_map.items():
+        print(mod, ":", code)
+    print("=" * 30)
 
     # print("%s is %scached." % (__file__, yes_no(is_cached(__file__))))
     # print(stat(__file__))
